# Copyright (C) 2018-2023 Mark McIntyre
import pickle
import tempfile
import pandas as pd
import os
# imports are required to load the picklefiles
try:
    from wmpl.Trajectory.CorrelateRMS import TrajectoryReduced, DatabaseJSON # noqa: F401
    from wmpl.Trajectory.CorrelateRMS import MeteorObsRMS, PlateparDummy, MeteorPointRMS # noqa: F401
except:
    print('WMPL not available')

from ukmon_meteortools.ukmondb.getLiveImages import _download


def getTrajPickle(trajname):
    """ Retrieve a the pickled trajectory for a matched detection 
    
    Arguments:  
        trajname:   [string] Name of the trajectory eg 20230502_025228.374_UK_BE

    Returns:
        WMPL traj object

    Notes:
        WMPL must be in the $PYTHONPATH for this function to be available

    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/pickle/getpickle'
    fmt = 'pickle'
    apicall = f'{apiurl}?reqval={trajname}&format={fmt}'
    matchpickle = pd.read_json(apicall, lines=True)
    if 'url' in matchpickle:
        outdir = tempfile.mkdtemp()
        _download(matchpickle.url[0], outdir, matchpickle.filename[0])

        localfile = os.path.join(outdir, matchpickle.filename[0])
        with open(localfile, 'rb') as f:
            traj = pickle.load(f, encoding='latin1')
        try:
            os.remove(localfile)
        except:
            print(f'unable to remove {localfile}')
    else:
        traj = None
    return traj



# Copyright (C) 2018-2023 Mark McIntyre
# 
#
import os
import glob
import xmltodict

from shapely.geometry import Polygon
from shapely.geometry import Point


def _munchKML(kmlFilename):
    """ Private function to work around circular dependency 
    """
    with open(kmlFilename) as fd:
        x = xmltodict.parse(fd.read())
        cname = x['kml']['Folder']['name']
        coords = x['kml']['Folder']['Placemark']['MultiGeometry']['Polygon']['outerBoundaryIs']['LinearRing']['coordinates']
        coords = coords.split('\n')
        ptsarr=[]
        for lin in coords:
            s = lin.split(',')
            ptsarr.append((float(s[0]), float(s[1])))
        polyg = Polygon(ptsarr)
        return cname, polyg 


def pointInsideFov(latDegs,lngDegs, kmlFilename):
    """
    Test if a point is inside the field of view of a KML file. Useful for testing which 
    cameras might have seen an event at a known location.  

    Arguments:  
        latDegs:        [degrees] latitude of the point  
        lngDegs:        [degrees] latitude of the point  
        kmlFileName:    [str] full path to the KML file to test  

    Returns:  
        True or False  
    """

    c1, p1 = _munchKML(kmlFilename)
    pt = Point(latDegs, lngDegs)
    return p1.contains(pt)


def checkKMLOverlap(kmfile1, kmfile2):
    """
    Test if two KML files overlap. Useful for finding which other cameras 
    might have seen an event seen by the first camera. 

    Arguments:  
        kmlfile1:       [str] full path to first KML file  
        kmlfile2:       [str] full path to second KML file  

    Returns:  
        True or False
    """
    _,p1 = _munchKML(kmfile1)
    _,p2 = _munchKML(kmfile2)
    return p1.intersects(p2)


def getOverlapWith(srcfolder, kmlpattern='*-25km.kml', refcam='UK0006'):
    """
    Check for overlap between the named camera, and every KML file in a folder, at the pattern altitude  

    Arguments:  
        srcfolder:  [str] path to folder containing KML files to test.   
        kmlpattern: [str] kml pattern to match. Default "*-25km.kml"  
        refcam:     [str] Cameraid to check against. Its KML must be in srcfolder  
     
    Returns:  
        list of camera IDs that overlap with the target  
    """
    kmllist = glob.glob1(srcfolder, kmlpattern)
    currmatches=[]
    refkml = f'{refcam}{kmlpattern[1:]}'
    currmatches.append(refcam[:6])
    print('checking ', refkml)
    for testkml in kmllist:
        if testkml != refkml:
            testcam,_ = os.path.splitext(testkml)
            if checkKMLOverlap(os.path.join(srcfolder, refkml), os.path.join(srcfolder, testkml)) is True:
                currmatches.append(testcam[:6])
    return currmatches



def getOverlappingCameras(srcfolder, kmlpattern='*-25km.kml'):
    """
    Check for all overlaps in the folder at the pattern altitude. Returns a massive 2d array  

    Arguments:  
        srcfolder:  [str] path to folder containing KML files to test.   
        kmlpattern: [str] kml pattern to match. Default "*-25km.kml"  

    Returns:  
        list of lists of groups of overlaping IDs 
     
    """
    kmllist = glob.glob1(srcfolder, kmlpattern)
    kmllist2 = kmllist
    matches = []

    for refkml in kmllist:
        currmatches=[]
        refcam,_ = os.path.splitext(refkml)
        currmatches.append(refcam[:6])
        for testkml in kmllist2:
            if testkml != refkml:
                testcam,_ = os.path.splitext(testkml)
                if checkKMLOverlap(os.path.join(srcfolder, refkml), os.path.join(srcfolder, testkml)) is True:
                    currmatches.append(testcam[:6])
        matches.append(currmatches)
    return matches



# Copyright (C) 2018-2023 Mark McIntyre
import os
import platform
import base64
import email
from email.mime.text import MIMEText

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build

# If modifying these scopes, delete the file token.json.
SCOPES =['https://mail.google.com/']


def _getGmailCreds(tokfile=None, crdfile=None):
    creds = None
    # The file token.json stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    if tokfile is None:
        tokfile = '~/.ssh/gmailtoken.json'
    if crdfile is None:
        crdfile = '~/.ssh/gmailcreds.json'
    tokfile = os.path.expanduser(tokfile)
    crdfile = os.path.expanduser(crdfile)
    if os.path.exists(tokfile):
        creds = Credentials.from_authorized_user_file(tokfile, SCOPES)
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            if not os.path.isfile(crdfile):
                print(f'to use this function you must have stored your Google OAUTH2 secret json file in {crdfile}')
                print('To set up OAUTH2 go to your google cloud console, select APIs then Credentials, and add an OAUTH2 desktop client ID.')
                return None
            flow = InstalledAppFlow.from_client_secrets_file(crdfile, SCOPES)
            creds = flow.run_local_server(port=0)
        # Save the credentials for the next run
        with open(tokfile, 'w') as token:
            token.write(creds.to_json())
    return creds


def _create_message(sender, to, subject, message_text):
    msg = MIMEText(message_text)
    msg['to'] = to
    msg['from'] = sender
    msg['subject'] = subject
    return {'raw': base64.urlsafe_b64encode(msg.as_string().encode('utf-8')).decode('utf-8')}



def sendAnEmail(mailrecip, message, msgtype, mailfrom, files=None, tokfile=None, crdfile=None):
    """ sends an email using gmail. 
    
        Arguments:  
            mailrecip:  [string] email address of recipient.  
            message:    [string] the message to send.  
            msgtype:    [string] Prefix for the subject line, eg Test, Warning.  
            mailfrom:   [string] email address of sender.   
            files:      [list]   list of files to attach, not currently implemented.  

        Returns:  
            Nothing, though a message is printed onscreen.  

        Notes:  
            You must have gmail OAUTH2 set up. The gmail credentials default to gmailtoken.json and
            gmailcreds.json in the  $HOME/.ssh folder.  
        """
    
    if msgtype is None:
        msgtype = platform.uname()[1]

    # email a summary to the mailrecip
    creds = _getGmailCreds(tokfile, crdfile)
    if not creds:
        return 
    service = build('gmail', 'v1', credentials=creds)

    subj ='{:s}: {:s}'.format(msgtype, message[:30])
    message = '{:s}: {:s}'.format(msgtype, message)

    mailmsg = _create_message(mailfrom, mailrecip, subj, message)

    try:
        retval = (service.users().messages().send(userId='me', body=mailmsg).execute())
        print('Message Id: %s' % retval['id'])
    except:
        print('An error occurred sending the message')


def forwardAnEmail(reciplist, msgid=None, tokfile=None, crdfile=None):
    """ Forward email from a gmail account to a list of recipients  
    
        Arguments:  
            recplist:   [list of strings] list of addresses to forward to  

        Keyword arguments:  
            msgid:      [string] gmail message id. If None, then all unread mail is forwarded  
            tokfile:    [string] full path to a gmail oauth2 json token file  
            crdfile:    [string] full path to a gmail oauth2 json credentials file for initial authorization 

        To obtain the oauth2 credentials file, go to the google cloud console, select APIs and enable gmail. Then 
         go to Credentials and create an OAUTH2 Client ID for Desktop and download the JSON credentials file. 
         Upon first run, you'll be taken to the gmail authorisation screen and the token file will be created. 
         Subsequent runs will use the token file. The creds file will be used to reauthorise periodically. 
       """
    creds = _getGmailCreds(tokfile, crdfile)
    if not creds:
        return 
    service = build('gmail', 'v1', credentials=creds)
    userid = 'me'
    if msgid is None:
        try:
            msglist = (service.users().messages().list(userId=userid, labelIds=['INBOX','UNREAD']).execute())
            for msg in msglist['messages']:
                msgid = msg['id']
                # retrieve raw mail
                message = (service.users().messages().get(userId='me', id=msgid, format='raw').execute())
                # decode it from base64
                decmsg = base64.urlsafe_b64decode(message['raw'])
                newmsg = email.message_from_bytes(decmsg)
                newmsg.add_header('Reply-To',newmsg['from']) 
                newmsg.replace_header('Subject','Fwd: ' + newmsg['subject'])
                # Send it
                for recip in reciplist:
                    newmsg.replace_header('To', recip)
                    mailmsg = {'raw': base64.urlsafe_b64encode(newmsg.as_string().encode('utf-8')).decode('utf-8')}
                    retval = (service.users().messages().send(userId='me', body=mailmsg).execute())
                    # This will mark the messagea as read
                    service.users().messages().modify(userId=userid, id=msgid, body={'removeLabelIds': ['UNREAD']}).execute() 
                    print(retval)
        except Exception:
            print('Nothing to forward')
    else:
        try:
            message = (service.users().messages().get(userId='me', id=msgid, format='raw').execute())
            # decode it from base64
            decmsg = base64.urlsafe_b64decode(message['raw'])
            newmsg = email.message_from_bytes(decmsg)
            newmsg.add_header('Reply-To',newmsg['from']) 
            newmsg.replace_header('Subject','Fwd: ' + newmsg['subject'])
            # Send it
            for recip in reciplist:
                newmsg.replace_header('To', recip)
                mailmsg = {'raw': base64.urlsafe_b64encode(newmsg.as_string().encode('utf-8')).decode('utf-8')}
                retval = (service.users().messages().send(userId='me', body=mailmsg).execute())
                # This will mark the messagea as read
                service.users().messages().modify(userId=userid, id=msgid, body={'removeLabelIds': ['UNREAD']}).execute() 
                print(retval)
        except Exception:
            print('Nothing to forward')
    return 



#
# Example showing how to use the matchdata API from Python
#
# Copyright (C) 2018-2023 Mark McIntyre

"""
Examples showing how to use the APIs. 
"""
import json
import requests 
import pandas as pd


def matchApiCall(reqtyp, reqval):
    """get names of all matches for a given date """
    apiurl = 'https://api.ukmeteornetwork.co.uk/matches'
    apicall = f'{apiurl}?reqtyp={reqtyp}&reqval={reqval}'
    matchlist = pd.read_json(apicall, lines=True)
    return matchlist


def getMatchPickle(patt):
    """
    get the pickled trajectory for a given matched event   
    Returns a json object containing the pickled data
    
    Example pattern 20230501_002536.754_UK
    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/pickle/getpickle'
    apicall = f'{apiurl}?reqval={patt}'
    res = requests.get(apicall, stream=True)
    if res.status_code == 200:
        data=b''
        for chunk in res.iter_content(chunk_size=4096):
            data = data + chunk
        jsd = json.loads(data)  
    else:
        jsd = None  
    return jsd


def detailApiCall1(reqtyp, reqval):
    """ get details of one matched event """
    apiurl = 'https://api.ukmeteornetwork.co.uk/matches'
    apicall = f'{apiurl}?reqtyp={reqtyp}&reqval={reqval}'
    evtdetail = pd.read_json(apicall, typ='series')
    return evtdetail


def detailApiCall2(reqtyp, matchlist):
    """
    get details for the events in the match list for a given date
    and put them in a pandas dataframe, then sort by brightest
    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/matches'
    details = []
    for id in matchlist.head(5).orbname:
        reqval = id
        apicall = f'{apiurl}?reqtyp={reqtyp}&reqval={reqval}'
        details.append(pd.read_json(apicall, typ='series'))
    df = pd.DataFrame(details)
    df = df.sort_values(by=['_mag'])
    return df


def getLiveimageList(dtstr):
    """ 
    Get a list URLs of livestream images matching a pattern YYYYMMDD_HHMMSS.  
    The seconds and minutes parts are optional but huge amounts of data may get returned.  

    Note that we only keep the last month of images and so this function won't return
    anything for older dates. 
    Note also that the URLs are presigned and valid only for five minutes.  

    Example pattern: '20230421_2122'
    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/liveimages/getlive'
    liveimgs = pd.read_json(f'{apiurl}?pattern={dtstr}')
    return liveimgs


def getFireballFiles(patt):
    """ 
    Get a list of URLs for the fireball data matching a pattern of the form UKxxxx_YYYYMMDD_HHMMSS
    Nothing will be returned if there is no fireball data available. The URLs are presigned and valid
    for five minutes. 

    Example pattern: 'UK0006_20230421_2122'

    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/fireballs/getfb'
    fbfiles = pd.read_json(f'{apiurl}?pattern={patt}')
    return fbfiles


def trajectoryAPI(trajname):
    """
    Returns a JSON object containing a WMPL trajectory object for a matched event.  

    Arguments:  
        trajname:   [string] trajectory name eg "20230502_025228.374_UK_BE"
    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/pickle/getpickle'
    fmt = 'json'
    apicall = f'{apiurl}?reqval={trajname}&format={fmt}'
    res = requests.get(apicall, stream=True)
    if res.status_code == 200:
        data=b''
        for chunk in res.iter_content(chunk_size=4096):
            data = data + chunk
        jsd = json.loads(data)  
    else:
        jsd = None  
    return jsd



# Copyright (C) 2018-2023 Mark McIntyre
# flake8: noqa

"""
The functions in this module require RMS and/or WMPL to be in the PythonPath.  

Require RMS:  
    multiDayRadiant  
    multiTrackStack  
    analyseUFOwithRMS  

Require WMPL:  
    multiEventGroundMap  
    plotCAMSOrbits  
    plotRMSOrbits  

"""

from .multiDayRadiant import multiDayRadiant
from .multiTrackStack import multiTrackStack
from .analyseUFOwithRMS import analyseUFOwithRMS
from .multiEventGroundMap import multiEventGroundMap
from .plotCAMSOrbits import plotCAMSOrbits
from .plotRMSOrbits import plotRMSOrbits



# Copyright (C) 2018-2023 Mark McIntyre
# flake8: noqa
"""
Utility functions used across the UKMON toolset

List of Functions:  
jd2Date, date2JD, datetime2JD, jd2DynamicalTimeJD, jd2LST, sollon2jd  

greatCircleDistance, angleBetweenSphericalCoords, calcApparentSiderealEarthRotation   
calcNutationComponents, equatorialCoordPrecession,  raDec2AltAz, altAz2RADec  
altAz2RADec_vect, raDec2AltAz_vect, equatorialCoordPrecession_vect  

annotateImage, annotateImageArbitrary  

getActiveShowers, getActiveShowersStr, getShowerDets, getShowerPeak  

sendAnEmail, forwardAnEmail

"""
from .Math import jd2Date, date2JD,datetime2JD, jd2DynamicalTimeJD, JULIAN_EPOCH, J2000_JD, jd2LST
from .Math import greatCircleDistance, angleBetweenSphericalCoords, calcApparentSiderealEarthRotation
from .Math import calcNutationComponents, equatorialCoordPrecession,  raDec2AltAz, altAz2RADec
from .Math import altAz2RADec_vect, raDec2AltAz_vect, equatorialCoordPrecession_vect
from .annotateImage import annotateImage, annotateImageArbitrary
from .convertSolLon import sollon2jd
from .getActiveShowers import getActiveShowers, getActiveShowersStr
from .getShowerDates import getShowerDets, getShowerPeak
from .sendAnEmail import sendAnEmail, forwardAnEmail
from .drawFTPfile import drawFTPFile
from .plotTrack import trackToDistvsHeight, trackToTimevsVelocity, trackToTimevsHeight
from .getOverlappingFovs import checkKMLOverlap, pointInsideFov, getOverlapWith, getOverlappingCameras

#from .findNearDuplicates import findNearDuplicates



# Copyright (C) 2018-2023 Mark McIntyre
# flake8: noqa
"""
Functions to load and manage various meteor file formats
- RMS FTPdetectInfo files
- RMS Platepar fiels
- The IMO Working shower list and full stream data list
- UFO Capture XML files
- UFO Analyser XML files

"""

from .ftpDetectInfo import filterFTPforSpecificTime, writeNewFTPFile, loadFTPDetectInfo, MeteorObservation
from .imoWorkingShowerList import IMOshowerList, majorlist, minorlist
from .platepar import loadPlatepars, platepar
from .UFOAnalyzerXML import UAXml
from .UFOCapXML import UCXml
from .kmlHandlers import trackCsvtoKML, trackKMLtoCsv, getTrackDetails, readCameraKML



# Copyright (C) 2018-2023 Mark McIntyre
import numpy as np
import requests
import json
import os 
import pandas as pd

from ukmon_meteortools.fileformats import trackCsvtoKML


def trajectoryKML(trajname, outdir=None):
    """
    Create a 3-d KML file showing the trajectory of a meteor which can be 
    loaded into Google Maps. 
    
    Arguments:  
        trajname:   [string] trajectory name eg 20230213_025913.678_UK
        outdir:     [string] where to save the file. Defaults to current location  

    Returns:  
        kmlfile:    name of the saved KML file  
    """
    apiurl = 'https://api.ukmeteornetwork.co.uk/pickle/getpickle'
    fmt = 'json'
    apicall = f'{apiurl}?reqval={trajname}&format={fmt}'
    res = requests.get(apicall, stream=True)
    if res.status_code == 200:
        data=b''
        for chunk in res.iter_content(chunk_size=4096):
            data = data + chunk
        jsd = json.loads(data)
        lats = []
        lons = []
        hts = [] 
        times = []
        for js in jsd['observations']:
            k = list(js.keys())[0]
            thisstat = js[k]
            lats += thisstat['model_lat']
            lons += thisstat['model_lon']
            hts += thisstat['model_ht']
            times += thisstat['time_data']
        df = pd.DataFrame({"lats": np.degrees(lats), "lons": np.degrees(lons), "alts": hts, "times": times})
        df = df.sort_values(by=['times', 'lats'])
        outfname = os.path.join(outdir, f'{trajname}.kml')
        if outdir is None:
            outdir = '.'
            os.makedirs(outdir, exist_ok=True)
        trackCsvtoKML(trajname, df, saveOutput=True, outdir=outdir)
        return outfname
    else:
        print('no match')
        return None



# Copyright (C) 2018-2023 Mark McIntyre

# use the API to retrieve single-station data

import datetime 
import requests
import pandas as pd


def getDetections(dtstr, interval='m'):
    """ 
    Returns a list of detections within one interval of the specified time  

    Arguments:  
        dtstr:      [string] Date to search for in the format YYYYMMDD_HHMMSS  
        interval:   [string] window to search around the specified time. Currently only 'm' supported  

    Returns:  
        pandas dataframe containing [ID, Dtstamp, Filename] of any matching detections  

    Note:  
        dtstr must be at least YYYYMMDD_HHMM. If seconds are omitted, the window wil be centred on 
        hh:mm:30.  

    """
    srchapi = 'https://api.ukmeteornetwork.co.uk/detections?'
    if len(dtstr) < 13:
        print('date range too wide. Must contain minutes')
        return False
    if len(dtstr) == 13:
        dtstr = dtstr + '30'
    if len(dtstr) == 14:
        dtstr = dtstr + '0'
    dt = datetime.datetime.strptime(dtstr[:15], '%Y%m%d_%H%M%S')
    if interval == 'm':
        dt1 = (dt + datetime.timedelta(seconds=-30)).strftime('%Y-%m-%dT%H:%M:%S.000Z')
        dt2 = (dt + datetime.timedelta(seconds=30)).strftime('%Y-%m-%dT%H:%M:%S.000Z')
    apiurl = f'{srchapi}a={dt1}&b={dt2}&op=t:S'

    res = requests.get(apiurl)
    ids = []
    filenames = []
    dtstamps = []
    if res.status_code == 200:
        rawdata=res.text.strip()
        rawdata = rawdata[8:-2]
        spls = rawdata.split('"')
        for spl in spls:
            if len(spl) > 10:
                thisrow = spl.split(',')
                ids.append(thisrow[4])
                fname = thisrow[5].split('/')[-1].strip()
                filenames.append(fname)
                dtstamps.append(fname[10:25])
        df = pd.DataFrame({'ID':ids, 'Dtstamp':dtstamps, 'Filename':filenames})
        return df
    else:
        return False



# Copyright (C) 2018-2023 Mark McIntyre
# flake8: noqa
"""
Functions to access data in the UKMON database.   

getECSVs        get the ECSV file or files for a given date-time  
getLiveImages   functions to get a list of livestream images or fireball imaqegs  
trajectoryKML   get a KML file for a UKMON trajectory, suitable for use in google maps  
trajPickle      get a WMPL pickle object for a trajectory  
apiExampleCode  various examples of how to use the APIs  
"""

from .ECSVhandler import getECSVs
from .apiExampleCode import matchApiCall, detailApiCall1, detailApiCall2, trajectoryAPI
from .apiExampleCode import getMatchPickle, getLiveimageList, getFireballFiles
from .getLiveImages import getLiveJpgs, getFBfiles, createTxtFile
from .trajectoryKML import trajectoryKML
from .trajPickle import getTrajPickle
from .getDetections import getDetections
